// start
package metaheuristicOptimization.algorithms.pso;

import metaheuristicOptimization.IObjectiveFunction;
import metaheuristicOptimization.Individual;
import metaheuristicOptimization.OptimizationAlgorithm;
import metaheuristicOptimization.searchSpaces.Rn.Rn;

/**
 * the Particle Swarm Optimization algorithm with global update
 * 
 * @param <X>
 *          the solution space
 */
// end
public class PSO<X> extends OptimizationAlgorithm<double[], X> { // start

  /** the population size */
  public int ps;

  /** the constant c */
  public double c;

  /** the constant d */
  public double d;

  /** the limits */
  public Rn rn;

  /** create */
  public PSO() {
    super();
    this.ps = 128;
    this.c = 0.5d;
    this.d = 0.5d;
  }

  /**
   * Solve the optimization problem
   * 
   * @param f
   *          the objective function
   * @return the best individual found
   */
  @Override
  @SuppressWarnings("unchecked")
  // end
  public Individual<double[], X> solve(final IObjectiveFunction<X> f) {
    final PSOIndividual<X>[] swarm;
    PSOIndividual<X> cur;
    Individual<double[], X> best;
    double limitV;
    int i, j;

    swarm = new PSOIndividual[this.ps];
    best = new Individual<>();
    best.g = new double[this.rn.dim];

    limitV = 0.1 * (this.rn.max - this.rn.min);

    for (i = swarm.length; (--i) >= 0;) {
      swarm[i] = cur = new PSOIndividual<>(this.nullary.create(this.random));
      cur.x = this.gpm.gpm(cur.g);
      cur.v = f.compute(cur.x);
      copyIndividual(cur.best, cur);
      if (cur.v < best.v) {
        copyIndividual(best, cur);
      }

      if (this.termination.shouldTerminate()) {
        return best;
      }
    }

    for (;;) {

      for (i = swarm.length; (--i) >= 0;) {
        cur = swarm[i];
        for (j = this.rn.dim; (--j) >= 0;) {
          cur.velocity[j] = Math.min(limitV, Math.max(-limitV,
              cur.velocity[j] = ((this.random.nextDouble() * this.c) * (cur.best.g[j] - cur.g[j]))
                  + ((this.random.nextDouble() * this.d) * (best.g[j] - cur.g[j]))));
        }
      }

      for (i = swarm.length; (--i) >= 0;) {
        cur = swarm[i];
        for (j = this.rn.dim; (--j) >= 0;) {
          cur.g[j] = Math.max(this.rn.min, Math.min(this.rn.max, cur.g[j] + cur.velocity[j]));
          cur.x = this.gpm.gpm(cur.g);
          cur.v = f.compute(cur.x);
          if (cur.v < cur.best.v) {
            copyIndividual(cur.best, cur);
            if (cur.v < best.v) {
              copyIndividual(best, cur);
            }
          }
          if (this.termination.shouldTerminate()) {
            return best;
          }
        }
      }
    }

  }// start

  /**
   * copy an individual
   * 
   * @param dest
   *          the destination
   * @param source
   *          the source
   */
  static <V> void copyIndividual(final Individual<double[], V> dest,
      final Individual<double[], V> source) {
    dest.v = source.v;
    System.arraycopy(source.g, 0, dest.g, 0, source.g.length);
    if (dest.x != dest.g) {
      dest.x = source.x;
    }
  }
}