package metaheuristicOptimization.searchSpaces.trees.data;

import java.util.Random;

/** The node type of a node has two main functions: 1) It describes which types of children
 * are allowed to occur at which position (if any), and 2) it is used to instantiate nodes
 * of the type. Notice that 1) is the basic requirement for strongly-typed GP: We can
 * specify which type of node can occur as child of which other node and thus, build a
 * complex type system and pre-define node structures precisely. Point 2) somehow
 * reproduces the capabilities of Java's reflection system with the extension of allowing
 * us to perform some additional, randomized actions. Matter of fact, with the class
 * ReflectionNodeType, we defer the node creation to Java's reflection mechanisms in cases
 * where no randomized instantiation actions are required.
 * @param <NT>
 *          the specific node type
 * @author Thomas Weise */
public class NodeType<NT extends Node<?>> {
  /** the key counter */
  static int keyCounter = 0;

  /** no children */
  public static final NodeTypeSet<?>[] EMPTY_CHILDREN = new NodeTypeSet[0];

  /** a list of possible node types for each child */
  private final NodeTypeSet<?>[] chTypes;

  /** a search key */
  final int key;

  /** Create a new node type record
   * @param ch
   *          the types of the possible chTypes */
  public NodeType(final NodeTypeSet<?>[] ch) {
    super();
    this.chTypes = (((ch != null) && ((ch.length) > 0)) ? ((NodeTypeSet[]) ch)
        : (NodeType.EMPTY_CHILDREN));
    this.key = (NodeType.keyCounter++);
  }

  /** Get the number of chTypes of this node type
   * @return the number of chTypes of this node type */
  public final int getChildCount() {
    return this.chTypes.length;
  }

  /** Get the possible types for the chTypes at the specific index.
   * @param index
   *          the child index
   * @return the possible types of that child */
  public final NodeTypeSet<?> getChildTypes(final int index) {
    return this.chTypes[index];
  }

  /** Instantiate a node
   * @param children
   *          a given set of children
   * @param r
   *          the randomizer
   * @return the new node */
  public NT instantiate(final Node<?>[] children, final Random r) {
    throw new UnsupportedOperationException();
  }

  /** Does this node type describe a terminal node?
   * @return true if this node type describes a terminal node, false otherwise */
  public final boolean isTerminal() {
    return (this.chTypes.length <= 0);
  }
}