package metaheuristicOptimization.searchSpaces.trees.ops;

import java.io.Serializable;
import java.util.Random;

import metaheuristicOptimization.searchSpaces.trees.data.Node;
import metaheuristicOptimization.searchSpaces.trees.data.NodeType;
import metaheuristicOptimization.searchSpaces.trees.data.NodeTypeSet;

/** A base class for tree-based search operations
 * @author Thomas Weise */
public class TreeOperation implements Serializable {
  /** a constant required by Java serialization */
  private static final long serialVersionUID = 1;

  /** the maximum depth */
  public final int maxDepth;

  /** Create a new tree operation
   * @param md
   *          the maximum tree depth */
  protected TreeOperation(final int md) {
    super();
    this.maxDepth = Math.max(2, md);
  }

  /** Create a sub-tree of the specified size maximum depth. This function facilitates
   * both, an implementation of the full method Algorithm 31.1 and one of the grow method
   * Algorithm 31.2, which can be chosen by setting the parameter 'full' apropriately. It
   * can be used to create trees during the random population initialization phase or
   * during mutation steps.
   * @param types
   *          the node types available for creating the tree
   * @param maxDepth
   *          the maximum depth of the tree
   * @param full
   *          Should we construct a sub-tree according to the full method? If full is
   *          false, grow is used.
   * @param r
   *          the random number generator
   * @return the new tree
   * @param <NT>
   *          the node type */
  @SuppressWarnings("unchecked") public static final <NT extends Node<?>> NT createTree(
      final NodeTypeSet<?> types, final int maxDepth, final boolean full, final Random r) {
    NodeType<?> t;
    NT[] x;
    int i;
    t = null;
    if (maxDepth <= 1) {
      t = types.randomTerminalType(r);
    } else {
      if (full) {
        t = types.randomNonTerminalType(r);
      }
      if (t == null) {
        t = types.randomType(r);
      }
    }
    if (t.isTerminal()) { return ((NT) (types.randomTerminalType(r).instantiate(null, r))); }
    i = t.getChildCount();
    x = ((NT[]) (new Node[i]));
    for (; (--i) >= 0;) {
      x[i] = TreeOperation.createTree(t.getChildTypes(i), maxDepth - 1, full, r);
    }
    return ((NT) (t.instantiate(x, r)));
  }
}